#!/usr/bin/env python2
# conding: utf-8

import os
import sys
import unittest
import logging

from pykit import shell


class TestCommand(unittest.TestCase):

    def setUp(self):
        self.out_buf = os.path.join(os.getcwd(), 'out_buf')
        self.backup_argv = sys.argv
        sys.argv = ['python']

        logging.basicConfig(level=logging.CRITICAL)

    def tearDown(self):
        sys.argv = self.backup_argv
        try:
            os.remove(self.out_buf)
        except EnvironmentError as e:
            sys.stderr.write(repr(e))

    def execute_test(self, arguments, argv, out_str, exit_code):

        sys.argv.extend(argv)

        backup_stderr = sys.stderr
        backup_stdout = sys.stdout

        with open(self.out_buf, 'w') as fw:
            sys.stderr = fw
            sys.stdout = fw

            try:
                shell.command(**arguments)
            except SystemExit as e:
                if len(e.args) > 0:
                    self.assertEqual(exit_code, e.args[0])

        sys.stderr = backup_stderr
        sys.stdout = backup_stdout

        with open(self.out_buf, 'r') as fr:
            s = fr.read()
            self.assertEqual(s, out_str)

        sys.argv = sys.argv[:1]

    def test_command_no_such_command(self):

        testcases = (
            (
                {'echo': lambda *x: sys.stderr.write(repr(x))},
                [],
                'No such command: ',
                2,
            ),
            (
                {'echo': lambda *x: sys.stderr.write(repr(x))},
                ['echoo'],
                'No such command: echoo',
                2,
            ),
            (
                {'call': 'not_callable'},
                ['call'],
                'No such command: call',
                2,
            )
        )

        for arguments, argv, out_str, exit_code in testcases:
            self.execute_test(arguments, argv, out_str, exit_code)

    def test_command_execute_error(self):

        testcases = (
            (
                {'divi': lambda x, y: int(x)/int(y)},
                ['divi'],
                "TypeError('<lambda>() takes exactly 2 arguments (0 given)',)",
                1,
            ),
            (
                {'mod': {
                    'mod_2': lambda x: int(x)%2,
                    },
                },
                ['mod', 'mod_2'],
                "TypeError('<lambda>() takes exactly 1 argument (0 given)',)",
                1,
            ),
            (
                {'divi': lambda x, y: int(x)/int(y)},
                ['divi', '7', '0'],
                "ZeroDivisionError('integer division or modulo by zero',)",
                1,
            ),
            (
                {'divi': lambda x, y: int(x)/int(y)},
                ['divi', 'string', 'number'],
                '''ValueError("invalid literal for int() with base 10: \'string\'",)''',
                1,
            ),
        )

        for arguments, argv, out_str, exit_code in testcases:
            self.execute_test(arguments, argv, out_str, exit_code)

    def test_command_execute_normal(self):

        testcases = (
            (
                {'echo_repr': lambda *x: sys.stderr.write(repr(x))},
                ['echo_repr', 'hello_world'],
                "('hello_world',)",
                0,
            ),
            (
                {'divi': lambda x, y: int(x)/int(y)},
                ['divi', '7', '3'],
                '',
                1,
            ),
            (
                {'mod': {
                    'mod_2':lambda x: int(x)%2,
                    }
                },
                ['mod', 'mod_2', '3'],
                '',
                1,
            ),
        )

        for arguments, argv, out_str, exit_code in testcases:
            self.execute_test(arguments, argv, out_str, exit_code)

    def test_command_help_message(self):

        commands = {
            'echo_repr': (
                lambda x: sys.stdout.write(repr(x)),
                        ('x', {'nargs': '+', 'help': 'just an input message'}),
                    ),

            'foo': {
                'bar': lambda: sys.stdout.write('bar'),

                'bob': {
                    'plus': (
                        lambda x, y: sys.stdout.write(str(x + y)),
                        ('x', {'type': int, 'help': 'an int is needed'}),
                        ('y', {'type': int, 'help': 'an int is needed'}),
                        ),
                },
            },

            '__add_help__': {
                ('echo_repr',)           : 'output what is input.',
                ('foo', 'bar',)          : 'print a "bar".',
                ('foo', 'bob', 'plus',)  : 'do addition operation with 2 numbers.',
            },

            '__description__': 'this is an example command.',
        }

        testcases = (
            (
                [],

                "usage: python [-h] {echo_repr,foo bar,foo bob plus} ...\n" +
                "python: error: too few arguments\n",

                2,

                'no command input to get help message',
            ),

            (
                ['-h'],

                "usage: python [-h] {echo_repr,foo bar,foo bob plus} ...\n" +
                "\n" +
                "this is an example command.\n" +
                "\n" +
                "positional arguments:\n" +
                "  {echo_repr,foo bar,foo bob plus}\n" +
                "                        command(s) to select ...\n" +
                "    echo_repr           output what is input.\n" +
                "    foo bar             print a \"bar\".\n" +
                "    foo bob plus        do addition operation with 2 numbers.\n" +
                "\n" +
                "optional arguments:\n" +
                "  -h, --help            show this help message and exit\n",

                0,

                'use -h to get help message',
            ),

            (
                ['echo_repr', '-h'],

                "usage: python echo_repr [-h] x [x ...]\n" +
                "\n" +
                "positional arguments:\n" +
                "  x           just an input message\n" +
                "\n" +
                "optional arguments:\n" +
                "  -h, --help  show this help message and exit\n",

                0,

                'use valid command and -h to get parameter help message',
            ),

            (
                ['foo', 'bar', '-h'],

                "usage: python foo bar [-h]\n" +
                "\n" +
                "optional arguments:\n" +
                "  -h, --help  show this help message and exit\n",

                0,

                'use valid command and -h to get parameter help message when no parameter setted.',
            ),

            (
                ['foo', 'bob', 'plus', '-h'],

                "usage: python foo bob plus [-h] x y\n" +
                "\n" +
                "positional arguments:\n" +
                "  x           an int is needed\n" +
                "  y           an int is needed\n" +
                "\n" +
                "optional arguments:\n" +
                "  -h, --help  show this help message and exit\n",

                0,

                'use valid command and -h to get parameter help message when many parameters setted.',
            ),

            (
                ['echo_repr', 'hello', 'world'],

                "['hello', 'world']",

                0,

                'run with help message setted.',
            ),

            (
                ['foo', 'bar'],

                "bar",

                0,

                'run with help message setted and no parameter help message setted.',
            ),

            (
                ['foo', 'bob', 'plus', "1", "2"],

                "3",

                0,

                'run with help message setted and many parameter help messages setted.',
            ),

            (
                ['foo', 'bar', 'error'],

                "usage: python [-h] {echo_repr,foo bar,foo bob plus} ...\n" +
                "python: error: unrecognized arguments: error\n",

                2,

                'error when use extra arguments.'
            ),

            (
                ['foo', 'bob', 'plus', "1", 'string'],

                "usage: python foo bob plus [-h] x y\n" +
                "python foo bob plus: error: argument y: invalid int value: 'string'\n",

                2,

                'error when arguments has invalid type as required.'
            ),

        )

        for argv, out_str, exit_code, msg in testcases:
            self.execute_test(commands, argv, out_str, exit_code)



#!/usr/bin/env python2.6
# coding: utf-8

import datetime
import time
import unittest

import pytz

from pykit import timeutil
from pykit import ututil

dd = ututil.dd

test_case = {
    'ts': {
        'day_accuracy': 1485216000,
        'hour_accuracy': 1485241200,
        'second_accuracy': 1485244319,
    },

    'format':             {
        'default':        'Tue, 24 Jan 2017 07:51:59 UTC',
        'iso':            '2017-01-24T07:51:59.000Z',
        'utc':            'Tue, 24 Jan 2017 07:51:59 UTC',
        'archive':        '20170124-07',
        'compact':        '20170124-075159',
        'daily':          '2017-01-24',
        'daily_compact':  '20170124',
        'mysql':          '2017-01-24 07:51:59',
        'nginxaccesslog': '24/Jan/2017:07:51:59',
        'nginxerrorlog':  '2017/01/24 07:51:59',
    },
}


class TestTimeutil(unittest.TestCase):

    def test_direct_format(self):

        for fmt_key, tm_str in test_case['format'].items():

            # parse
            dt_key = timeutil.parse(tm_str, fmt_key)
            dt_direct = timeutil.parse(tm_str, timeutil.formats[fmt_key])

            self.assertTrue(dt_key == dt_direct)

            # format
            self.assertEqual(tm_str, timeutil.format(dt_key, fmt_key))
            self.assertEqual(tm_str,
                             timeutil.format(dt_key,
                                             timeutil.formats[fmt_key]))

            # format_ts
            now = int(time.time())
            self.assertEqual(timeutil.format_ts(now, fmt_key),
                             timeutil.format_ts(now, timeutil.formats[fmt_key]))

    def test_parse(self):

        for fmt_key, tm_str in test_case['format'].items():

            dt = timeutil.parse(tm_str, fmt_key)
            ts = timeutil.utc_datetime_to_ts(dt)

            if fmt_key == 'archive':
                self.assertEqual(test_case['ts']['hour_accuracy'], ts)
            elif fmt_key.startswith('daily'):
                self.assertEqual(test_case['ts']['day_accuracy'], ts)
            else:
                self.assertEqual(test_case['ts']['second_accuracy'], ts)

    def test_format(self):

        dt = timeutil.ts_to_datetime(test_case['ts']['second_accuracy'])

        for fmt_key, tm_str in test_case['format'].items():

            convert_tm_str = timeutil.format(dt, fmt_key)

            self.assertEqual(tm_str, convert_tm_str)

    def test_format_ts(self):

        for fmt_key, tm_str in test_case['format'].items():

            convert_tm_str = \
                timeutil.format_ts(test_case['ts']['second_accuracy'], fmt_key)

            self.assertEqual(tm_str, convert_tm_str)

    def test_ts_and_datetime_conversion(self):

        ts = timeutil.ts()

        dt = timeutil.ts_to_datetime(ts)
        converted_ts = timeutil.utc_datetime_to_ts(dt)

        self.assertEqual(ts, converted_ts)

    def test_timestamp(self):

        cases = [
            (timeutil.ts, 10, 1, 2),
            (timeutil.ms, 13, 0.001, 3),
            (timeutil.us, 16, 0.000001, 300),
            (timeutil.ns, 19, 0.000000001, 30000),
        ]

        for case in cases:
            timestamp_func, length, unit_ts, tolerance_ts = case

            ts1 = timestamp_func()

            time.sleep(unit_ts)

            ts2 = timestamp_func()

            self.assertTrue(ts1 < ts2 and ts2 < ts1 + tolerance_ts)

            self.assertEqual(length, len(str(ts2)))

    def test_convert_to_ts(self):

        ts = timeutil.ts()

        ms = ts * 1000
        us = ts * 1000000
        ns = ts * 1000000000

        self.assertEqual(ts, timeutil.ms_to_ts(ms))
        self.assertEqual(ts, timeutil.us_to_ts(us))
        self.assertEqual(ts, timeutil.ns_to_ts(ns))

    def test_to_ts(self):

        ts = timeutil.ts()

        cases = (
            ts,
            ts + 0.1,
            ts * 1000,
            ts * 1000 + 1,
            ts * 1000 + 0.1,

            ts * 1000 * 1000,
            ts * 1000 * 1000 + 1,
            ts * 1000 * 1000 + 0.1,

            ts * 1000 * 1000 * 1000,
            ts * 1000 * 1000 * 1000 + 1,
            ts * 1000 * 1000 * 1000 + 0.1,
        )

        for inp in cases:
            dd(inp, ts)

            self.assertEqual(ts, timeutil.to_sec(inp),
                             'convert {inp} to second'.format(inp=repr(inp)))

            self.assertEqual(ts, timeutil.to_sec(str(inp)),
                             'convert {inp} to second'.format(inp=repr(inp)))

    def test_to_ts_invalid_input(self):

        cases = (
            'a',
            '1',
            '1.1',
            -123456789,
            {},
            [],
            (),
            True,
            datetime.datetime.now(),
        )

        for inp in cases:

            with self.assertRaises(ValueError):
                timeutil.to_sec(inp)

    def test_is_timestamp(self):
        cases = (
            (False,                None, False,),
            (0,                    None, False,),
            ('0',                  None, False,),
            (u'0',                 None, False,),
            ((),                   None, False,),
            ([],                   None, False,),
            ({},                   None, False,),
            (type,                 None, False,),
            (149361634,            None, False,),
            (14936163419,          None, False,),
            (149361634100,         None, False,),
            (14936163410009,       None, False,),
            (149361634100011,      None, False,),
            (14936163410001119,    None, False,),
            (149361634100011122,   None, False,),
            (14936163410001112229, None, False,),
            (1493616341,           None, True,),
            (1493616341000,        None, True,),
            (1493616341000111,     None, True,),
            (1493616341000111222,  None, True,),
            (1493616341,           's',  True,),
            (1493616341,           'ms', False,),
            (1493616341,           'us', False,),
            (1493616341,           'ns', False,),
            (1493616341000,        's',  False,),
            (1493616341000,        'ms', True,),
            (1493616341000,        'us', False,),
            (1493616341000,        'ns', False,),
            (1493616341000111,     's',  False,),
            (1493616341000111,     'ms', False,),
            (1493616341000111,     'us', True,),
            (1493616341000111,     'ns', False,),
            (1493616341000111222,  's',  False,),
            (1493616341000111222,  'ms', False,),
            (1493616341000111222,  'us', False,),
            (1493616341000111222,  'ns', True,),
        )

        for s, unit, expected in cases:

            dd(s, unit, expected)

            rst = timeutil.is_timestamp(s, unit=unit)
            dd('rst: ', rst)

            self.assertEqual(expected, rst)

            # test input as string
            rst = timeutil.is_timestamp(str(s), unit=unit)
            dd('rst(str): ', rst)

            self.assertEqual(expected, rst)

    def test_datetime_to_ts(self):
        ts = time.time()

        dt = datetime.datetime.fromtimestamp(ts)
        r = timeutil.datetime_to_ts(dt)
        self.assertEqual(ts, r)

        test_timezones = (
            'US/Pacific',
            'Europe/Warsaw',
            'Asia/Shanghai',
        )

        for timezone_name in test_timezones:
            dt = datetime.datetime.fromtimestamp(
                ts, tz=pytz.timezone(timezone_name))

            r = timeutil.datetime_to_ts(dt)
            self.assertEqual(ts, r)

    def test_parse_with_timezone(self):
        cases = (
            ('2018-04-03 17:45:01', 'Asia/Shanghai', 1522748701),
            ('2018-04-03 17:45:01', 'UTC', 1522748701 + 3600 * 8),
        )

        for time_str, timezone, exp_ts in cases:
            dt = timeutil.parse(time_str, 'mysql', timezone=timezone)
            ts = timeutil.datetime_to_ts(dt)

            self.assertEqual(exp_ts, ts)



from pykit.p3json.test import PyTest


class TestDefault(object):
    def test_default(self):
        self.assertEqual(
            self.dumps(type, default=repr),
            self.dumps(repr(type)))


class TestPyDefault(TestDefault, PyTest): pass



from cStringIO import StringIO
from pykit.p3json.test import PyTest


class TestDump(object):
    def test_dump(self):
        sio = StringIO()
        self.json.dump({}, sio)
        self.assertEqual(sio.getvalue(), '{}')

    def test_dumps(self):
        self.assertEqual(self.dumps({}), '{}')

    def test_encode_truefalse(self):
        self.assertEqual(self.dumps(
                 {True: False, False: True}, sort_keys=True),
                 '{"false": true, "true": false}')
        self.assertEqual(self.dumps(
                {2: 3.0, 4.0: 5L, False: 1, 6L: True}, sort_keys=True),
                '{"false": 1, "2": 3.0, "4.0": 5, "6": true}')

    # Issue 16228: Crash on encoding resized list
    def test_encode_mutated(self):
        a = [object()] * 10
        def crasher(obj):
            del a[-1]
        self.assertEqual(self.dumps(a, default=crasher),
                 '[null, null, null, null, null]')


class TestPyDump(TestDump, PyTest): pass



from .accessor import (
    KeyValue,
    Value,
)

from .exceptions import (
    Aborted,
    ConnectionLoss,
    Deadlock,
    HigherTXApplied,
    RetriableError,
    TXError,
    TXTimeout,
    NotLocked,
    UnlockNotAllowed,
    UserAborted,
)

from .status import (
    ABORTED,
    COMMITTED,
    PURGED,

    STATUS,
)

from .storage import (
    StorageHelper,
    Storage,
)

from .zkstorage import (
    ZKStorage,
)

from .zkaccessor import (
    ZKKeyValue,
    ZKValue,
)

from .zktx import (
    TXRecord,
    ZKTransaction,

    list_recoverable,
    run_tx,

)

from .redisstorage import (
    RedisStorage,
)

from .redisaccessor import (
    RedisKeyValue,
    RedisValue,
)

from .slave import (
    Slave,
)

__all__ = [
    "KeyValue",
    "Value",

    "Aborted",
    "ConnectionLoss",
    "Deadlock",
    "HigherTXApplied",
    "RetriableError",
    "TXError",
    "TXTimeout",
    "UserAborted",
    "NotLocked",
    "UnlockNotAllowed",


    "ABORTED",
    "COMMITTED",
    "PURGED",

    "STATUS",

    "StorageHelper",
    "Storage",

    "ZKKeyValue",
    "ZKValue",

    "ZKStorage",

    "TXRecord",
    "ZKTransaction",

    "list_recoverable",
    "run_tx",

    "RedisStorage",

    "RedisKeyValue",
    "RedisValue",

    "Slave",
]



from .rangeset import (
    IntIncRange,
    IntIncRangeSet,
    Range,
    RangeDict,
    RangeSet,
    ValueRange,

    substract_range,

    intersect,
    substract,
    union,
)

__all__ = [
    "IntIncRange",
    "IntIncRangeSet",
    "Range",
    "RangeDict",
    "RangeSet",
    "ValueRange",

    "substract_range",

    "intersect",
    "substract",
    "union",
]



import unittest

from kazoo.client import KazooClient

from pykit import utdocker
from pykit import ututil
from pykit import zkutil

dd = ututil.dd

zk_tag = 'daocloud.io/zookeeper:3.4.10'
zk_name = 'zk_test'


class ZKTestBase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        utdocker.pull_image(zk_tag)

    def setUp(self):

        utdocker.create_network()
        utdocker.start_container(
            zk_name,
            zk_tag,
            env={
                "ZOO_MY_ID": 1,
                "ZOO_SERVERS": "server.1=0.0.0.0:2888:3888",
            },
            port_bindings={2181: 2181}
        )

        self.zk = KazooClient('127.0.0.1:2181')
        self.zk.start()

        self.zkauthed, _ = zkutil.kazoo_client_ext(
            {'hosts': '127.0.0.1:2181', 'auth': ('digest', 'xp', '123'),
             'acl': (('xp', '123', 'cdrwa'), ('foo', 'bar', 'rw'))})

        dd('start zk-test in docker')

    def tearDown(self):

        self.zk.stop()
        self.zkauthed.stop()
        utdocker.remove_container(zk_name)



#!/usr/bin/env python2
# coding: utf-8

# utfjson use python3 json module in python2.
#
# json module in python2 does not support disabling decoding string into
# unicode when json.loads().
#
# Thus if we do not know encoding of the json string we just can't load it.
# But the reality is that user input can be in any encoding or the input
# string might be an arbitrary byte array.
#
# In this case we need to laod json without decoding.
#
# Thus we ported python3 json module here and have made modification in order
# to let it pass python2 json test suites.
from pykit import p3json


# expected behavior to dump '我':
#           source             encoding=None  encoding='utf-8'
# unicode  u'\u6211'           '"\\u6211"'    '"\xe6\x88\x91"'
# str       '\xe6\x88\x91'     TypeError      '"\xe6\x88\x91"'


# ensure_ascii behavior to dump '我', in p3json:
#
#           source             ensure_ascii=True         ensure_ascii=False
# unicode  u'\u6211'           '"\\u6211"'               u'"\u6211"'
# str       '\xe6\x88\x91'     '"\\u00e6\\u0088\\u0091"' '"\xe6\x88\x91"'


def dump(obj, encoding='utf-8'):

    # - Using non-unicode with ensure_ascii=True results in unexpected output:
    #       p3fjson.dump('我',  ensure_ascii=True)  --> '"\\u00e6\\u0088\\u0091"'
    #
    #   We do not allow this to happen.
    #
    #
    # - Using nonunicode with ensure_ascii=False results in unexpected output:
    #       p3json.dump(u'我', ensure_ascii=False) --> u'"\u6211"'
    #
    #   It is rare to use unicode string in a serialized string.
    #   When transmitting object over network what we need is a byte
    #   series(str).
    #   Thus if input is unicode and ecoding is specified, we convert unicode
    #   to a utf-8 string first.

    if encoding is None:
        # If do not encode, source string must be unicode.
        ensure_unicode(obj)
        ensure_ascii = True
    else:
        obj = encode_unicode(obj, encoding)
        # We do not need to convert non-ascii chars in a encoded string.
        ensure_ascii = False

    return p3json.dumps(obj, ensure_ascii=ensure_ascii)


def load(s, encoding=None):

    # By default, do not try to decode non-unicode string in result.
    # There could be invalid chars.

    if s is None:
        return None

    rst = p3json.loads(s)

    if encoding is not None:
        rst = decode(rst, encoding)

    return rst


def ensure_unicode(o):

    if isinstance(o, str):
        raise TypeError('string({o} {tp}) must be unicode'
                        ' if ensure_ascii is True'.format(o=o, tp=type(o)))

    if isinstance(o, dict):
        for k, v in o.items():
            ensure_unicode(k)
            ensure_unicode(v)

    elif isinstance(o, (list, tuple)):
        for v in o:
            ensure_unicode(v)


def encode_unicode(o, encoding):

    if isinstance(o, unicode):
        return o.encode(encoding)

    if isinstance(o, dict):
        rst = {}
        for k, v in o.items():
            rst[encode_unicode(k, encoding)] = encode_unicode(v, encoding)

    elif isinstance(o, (list, tuple)):
        rst = []
        for v in o:
            rst.append(encode_unicode(v, encoding))
    else:
        rst = o

    return rst


def decode(o, encoding):

    if isinstance(o, str):
        return o.decode(encoding)

    if isinstance(o, dict):
        rst = {}
        for k, v in o.items():
            rst[decode(k, encoding)] = decode(v, encoding)

    elif isinstance(o, list):
        rst = []
        for v in o:
            rst.append(decode(v, encoding))
    else:
        rst = o

    return rst



from .humannum import (
    K,
    M,
    G,
    T,
    P,
    E,
    Z,
    Y,
    humannum,
    parsenum,
    parseint,

    value_to_unit,
    unit_to_value,
)

__all__ = [
    'K',
    'M',
    'G',
    'T',
    'P',
    'E',
    'Z',
    'Y',
    'humannum',
    'parsenum',
    'parseint',

    'value_to_unit',
    'unit_to_value',
]



from .dictutil import (
    add,
    addto,
    attrdict,
    attrdict_copy,
    breadth_iter,
    combine,
    combineto,
    contains,
    depth_iter,
    get,
    make_getter,
    make_getter_str,
    make_setter,
    subdict,

    AttrDict,
    AttrDictCopy,
)

__all__ = [
    "add",
    "addto",
    "attrdict",
    "attrdict_copy",
    "breadth_iter",
    "combine",
    "combineto",
    "contains",
    "depth_iter",
    "get",
    "make_getter",
    "make_getter_str",
    "make_setter",
    "subdict",

    "AttrDict",
    "AttrDictCopy",
]



from .strutil import (
    common_prefix,
    format_line,
    line_pad,
    tokenize,
    break_line,

    ColoredString,
    colorize,
    struct_repr,
    format_table,
    utf8str,
)

from .strutil import (
    blue,
    cyan,
    green,
    purple,
    red,
    white,
    yellow,

    optimal,
    normal,
    loaded,
    warn,
    danger,

    fading_color,
)

from .trie import (
    TrieNode,
    make_trie,
    sharding,
)

__all__ = [
    'common_prefix',
    'format_line',
    'line_pad',
    'tokenize',
    'break_line',

    'ColoredString',
    'colorize',

    'blue',
    'cyan',
    'green',
    'purple',
    'red',
    'white',
    'yellow',

    'optimal',
    'normal',
    'loaded',
    'warn',
    'danger',

    'fading_color',

    'struct_repr',
    'format_table',
    'utf8str',

    'TrieNode',
    'make_trie',
    'sharding',
]



#! /usr/bin/env python2
# coding: utf-8

import re
import yaml

from yaml import MappingNode
from yaml import ScalarNode


def load(s, encoding='utf-8'):

    # override to set output encoding
    def construct_yaml_str(loader, node):
        value = loader.construct_scalar(node)

        if encoding is not None:
            return value.encode(encoding)

        return value

    yaml.add_constructor(u'tag:yaml.org,2002:str', construct_yaml_str)

    return yaml.load(s)


def dump(s, encoding='utf-8', save_unicode=False):

    yaml.add_representer(str, represent_str)
    yaml.add_representer(dict, represent_mapping)

    # dump unicode as a str to remove '!!python/unicode' tag
    if not save_unicode:
        tag = u'tag:yaml.org,2002:str'
    else:
        tag = u'tag:yaml.org,2002:python/unicode'

    yaml.add_representer(unicode, lambda dumper,
                         data: dumper.represent_scalar(tag, data))

    if encoding is not None:
        dumped = yaml.dump(s, allow_unicode=True, encoding=encoding)

    else:
        dumped = yaml.dump(s, allow_unicode=True, encoding='utf-8')
        # unify encoding
        dumped = unicode(dumped, 'utf-8')

    # '...' is not necessary for yaml
    pattern = '\.\.\.\n$'

    return re.sub(pattern, '', dumped, 1)


def represent_str(dumper, data):

    # override to remove tag '!!python/str'
    tag = u'tag:yaml.org,2002:str'
    try:
        data = unicode(data, 'utf-8')
        style = None
    except UnicodeDecodeError:
        data = data.encode('base64')
        tag = u'tag:yaml.org,2002:binary'
        style = '|'

    return dumper.represent_scalar(tag, data, style=style)


def represent_mapping(dumper, mapping, flow_style=None):

    # override to dump map with item in any types
    value = []
    tag = u'tag:yaml.org,2002:map'
    node = MappingNode(tag, value, flow_style=flow_style)

    if dumper.alias_key is not None:
        dumper.represented_objects[dumper.alias_key] = node

    # there is the change
    # ex: [u'hello', '中国'].sort() raises an error

    # if hasattr(mapping, 'items'):
    #     mapping = mapping.items()
    #     mapping.sort()

    best_style = True
    for item_key, item_value in mapping.items():

        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)

        if not (isinstance(node_key, ScalarNode) and not node_key.style):
            best_style = False

        if not (isinstance(node_value, ScalarNode) and not node_value.style):
            best_style = False

        value.append((node_key, node_value))

    if flow_style is None:
        if dumper.default_flow_style is not None:
            node.flow_style = dumper.default_flow_style
        else:
            node.flow_style = best_style

    return node



import logging
import os
import threading
import time
from collections import defaultdict

import redis

from pykit import utfjson

logger = logging.getLogger(__name__)

# redis is thread safe

# NOTE: fork may duplicate file descriptor that confuses connection pool.
_pid_client = defaultdict(dict)
_lock = threading.RLock()


def get_client(ip_port):

    ip_port = normalize_ip_port(ip_port)

    pid = os.getpid()

    with _lock:
        o = _pid_client[ip_port]

        if pid not in o:
            o[pid] = redis.StrictRedis(*ip_port)

    return _pid_client[ip_port][pid]


def wait_serve(ip_port, timeout=5):

    t = time.time() + timeout

    rcl = get_client(ip_port)

    while time.time() < t:
        try:
            rcl.hget('foo', 'foo')
            logger.info('redis is ready: ' + repr(ip_port))
            return

        except redis.ConnectionError as e:
            logger.info('can not connect to redis: ' +
                        repr(ip_port) + ' ' + repr(e))
            time.sleep(0.1)
            continue
    else:
        logger.error('can not connect to redis: ' + repr(ip_port))
        raise


class RedisChannel(object):

    '''
    send message `data` through `channel`.
    `channel` is a list in redis.
    `peer` is "client" or "server".

    send is a rpush operation that adds an item to the end of a list.
    recv is a lpop operation that pops an item from the start of a list.
    '''

    other_peer = {
        'client': 'server',
        'server': 'client',
    }

    def __init__(self, ip_port, channel, peer, timeout=None):

        assert peer in self.other_peer

        self.ip_port = normalize_ip_port(ip_port)
        self.rcl = get_client(self.ip_port)

        # convert ['a', 'b'] to '/a/b'
        if isinstance(channel, (list, tuple)):
            channel = '/' + '/'.join(channel)

        self.channel = channel
        self.peer = peer.lower()
        self.send_list_name = '/'.join([self.channel, self.peer])
        self.recv_list_name = '/'.join([self.channel,
                                        self.other_peer[self.peer]])
        self.timeout = timeout

    def send_msg(self, data):
        j = utfjson.dump(data)
        self.rcl.rpush(self.send_list_name, j)

        if self.timeout is not None:
            self.rcl.expire(self.send_list_name, self.timeout)

    def recv_msg(self):

        v = self.rcl.lpop(self.recv_list_name)
        if v is None:
            return None

        return utfjson.load(v)

    def brecv_msg(self, timeout=0):

        v = self.rcl.blpop(self.recv_list_name, timeout=timeout)
        if v is None or len(v) != 2:
            return None

        # v is a tuple, (key, value)
        _, value = v

        return utfjson.load(value)

    def recv_last_msg(self):

        last = None
        while True:
            v = self.rcl.lpop(self.recv_list_name)
            if v is None:
                return utfjson.load(last)

            last = v

    def brecv_last_msg(self, timeout=0):

        msg = self.brecv_msg(timeout=timeout)
        if msg is None:
            return None

        last = self.recv_last_msg() or msg

        return last

    def peek_msg(self):
        v = self.rcl.lindex(self.recv_list_name, 0)
        if v is None:
            return None

        return utfjson.load(v)

    def rpeek_msg(self):
        v = self.rcl.lindex(self.recv_list_name, -1)
        if v is None:
            return None

        return utfjson.load(v)

    def list_channel(self, prefix):
        if isinstance(prefix, (list, tuple)):
            _prefix = '/' + '/'.join(prefix) + '/'
        else:
            _prefix = prefix

        if not _prefix.startswith('/'):
            raise ValueError(
                'prefix must starts with "/", but:' + repr(prefix))

        if _prefix.endswith('*'):
            raise ValueError(
                'prefix must NOT ends with "*", but:' + repr(prefix))

        if not _prefix.endswith('/'):
            raise ValueError('prefix must ends with "/", but:' + repr(prefix))

        _prefix = _prefix + '*'
        channels = self.rcl.keys(_prefix)

        rst = []
        for c in channels:

            for k in self.other_peer:
                k = '/' + k
                if c.endswith(k):
                    c = c[:-len(k)]
                    break
            else:
                logger.info('not a channel: ' + repr(c))
                continue

            if c not in rst:
                rst.append(c)

        return sorted(rst)


def normalize_ip_port(ip_port):
    if isinstance(ip_port, (int, long)):
        ip_port = ('127.0.0.1', ip_port)

    return ip_port





int_types = (int, long)
list_like = (type([]), type(()))

compatible_types = {
    type(None): (type(None), ),
    int:        (type(None), int, ),
    long:       (type(None), int,   long, ),
    float:      (type(None), int,   long, float, ),
    str:        (type(None), str,),
    tuple:      (type(None), tuple, ),
    list:       (type(None), list,),
}


class RangeException(Exception):
    pass


class Unmergeable(RangeException):
    pass


class ValueRange(list):

    def __init__(self, left, right, val=None):

        assert_compatible(left, right)

        if cmp_boundary(left, right) > 0:
            raise ValueError('left not smaller or equal right: {left}, {right}'.format(
                left=left, right=right))

        super(ValueRange, self).__init__([left, right, val])

    def cmp(self, b):

        # if a and b can be merged into one range, we say a == b
        #
        # Different from Range: adjacent ValueRange-s are not equal,
        # because they have their own value bound and can not be merged.

        if cmp_boundary(self[0], b[1]) >= 0:
            return 1

        if cmp_boundary(b[0], self[1]) >= 0:
            return -1

        # incomparable: overlapping or adjacent ranges
        return 0

    def intersect(self, b):

        if self.cmp(b) != 0:
            return None

        rst = [None, None] + self[2:]

        if self.cmp_left(b[0]) < 0:
            rst[0] = b[0]
        else:
            rst[0] = self[0]

        if self.cmp_right(b[1]) > 0:
            rst[1] = b[1]
        else:
            rst[1] = self[1]

        if rst[0] is not None and rst[0] == rst[1]:
            return None

        return rst

    def cmp_left(self, pos):
        # left is None means it is negative infinite
        return cmp_val(self[0], pos, none_cmp_finite=-1)

    def cmp_right(self, pos):
        # right is None means it is positive infinite
        return cmp_val(self[1], pos, none_cmp_finite=1)

    def is_adjacent(self, b):
        return cmp_boundary(b[0], self[1]) == 0

    def has(self, pos):
        return (self.cmp_left(pos) <= 0
                and self.cmp_right(pos) > 0)

    def length(self):
        if self[0] is None or self[1] is None:
            return float('inf')

        if isinstance(self[0], basestring):

            a, b = self[:2]
            l = max([len(a), len(b)])

            rst = 0.0
            ratio = 1.0
            for i in range(l):

                if i >= len(a):
                    va = 0.0
                else:
                    va = ord(a[i]) + 1.0

                if i >= len(b):
                    vb = 0.0
                else:
                    vb = ord(b[i]) + 1.0
                rst += (vb - va) / 257.0 * ratio
                ratio /= 257.0

            return rst
        else:
            return self[1] - self[0]

    def dup(self):
        return self.__class__(*self)

    def next_left(self):
        return self[1]

    def prev_right(self):
        return self[0]


class Range(ValueRange):

    def __init__(self, left, right):

        assert_compatible(left, right)

        if cmp_boundary(left, right) > 0:
            raise ValueError('left not smaller or equal right: {left}, {right}'.format(
                left=left, right=right))

        super(ValueRange, self).__init__([left, right])

    def cmp(self, b):

        # if a and b can be merged into one range, we say a == b
        #
        # Different from ValueRange: adjacent Range-s are equal,
        # because they can be merged into one.

        if cmp_boundary(self[0], b[1]) > 0:
            return 1

        if cmp_boundary(b[0], self[1]) > 0:
            return -1

        # incomparable: overlapping or adjacent ranges
        return 0


class IntIncRange(Range):

    def __init__(self, left, right):

        if left is not None and type(left) not in int_types:
            raise TypeError('{l} {ltyp} is not int or None'.format(
                l=left, ltyp=type(left)))

        if right is not None and type(right) not in int_types:
            raise TypeError('{r} {rtyp} is not int or None'.format(
                r=right, rtyp=type(right)))

        if cmp_boundary(left, right) > 0:
            raise ValueError('left not smaller or equal right: {left}, {right}'.format(
                left=left, right=right))

        list.__init__(self, [left, right])

    def cmp(self, b):

        # if a and b can be merged into one range, we say a == b

        if None not in (self[0], b[1]) and self[0] - b[1] > 1:
            return 1

        if None not in (b[0], self[1]) and b[0] - self[1] > 1:
            return -1

        # incomparable: overlapping or adjacent ranges
        return 0

    def is_adjacent(self, b):
        return (None not in (b[0], self[1])
                and self[1] + 1 == b[0])

    def has(self, pos):
        return (cmp_val(self[0], pos, none_cmp_finite=-1) <= 0
                and cmp_val(pos, self[1], none_cmp_finite=1) <= 0
                and type(pos) in int_types)

    def length(self):
        if None in self:
            return float('inf')

        return self[1] - self[0] + 1

    def next_left(self):
        return self[1] + 1

    def prev_right(self):
        return self[0] - 1


class RangeDict(list):

    default_range_clz = ValueRange

    def __init__(self, iterable=None, range_clz=None):

        if iterable is None:
            iterable = []

        self.range_clz = range_clz or self.default_range_clz

        super(RangeDict, self).__init__([self.range_clz(*x)
                                         for x in iterable])

        for i in range(0, len(self) - 1):
            if self[i].cmp(self[i + 1]) != -1:
                raise ValueError('range[{i}] {ri} does not smaller than range[{j}] {ripp}'.format(
                    i=i,
                    j=i + 1,
                    ri=self[i],
                    ripp=self[i + 1],
                ))

    def add(self, rng, val=None):

        rng = _to_range(self.range_clz, list(rng) + [val])

        i = bisect_left(self, rng)

        while i < len(self):

            if rng.cmp(self[i]) == 0:

                l, r = substract_range(self[i], rng)
                if l is None:
                    if r is None:
                        self.pop(i)
                    else:
                        self[i] = r
                        i += 1
                else:
                    if r is None:
                        self[i] = l
                        i += 1
                    else:
                        self.pop(i)
                        self.insert(i, r)
                        self.insert(i, l)
                        i += 2
            else:
                break

        if rng[0] is None:
            self.insert(0, rng)
        else:
            for i in range(len(self)):
                if self[i].cmp_left(rng[0]) > 0:
                    self.insert(i, rng)
                    break
            else:
                self.append(rng)

        self.normalize()

    def get(self, pos):
        rng = [pos, None]
        i = bisect_left(self, rng)

        if i == len(self) or not self[i].has(pos):
            raise KeyError('not in range: ' + repr(pos))

        return self[i][2]

    def get_min(self, is_lt=None):

        if len(self) == 0:
            raise ValueError("range dict is empty")

        if is_lt is None:
            is_lt = lambda a, b: a < b

        min_val_idx = 0
        for i in range(1, len(self)):
            if is_lt(self[i][2], self[min_val_idx][2]):
                min_val_idx = i

        min_val_rng = self[min_val_idx]

        return min_val_idx, min_val_rng, min_val_rng[2]

    def has(self, pos):
        rng = [pos, None]
        i = bisect_left(self, rng)

        if i == len(self):
            return False

        return self[i].has(pos)

    def length(self):
        rst = 0
        for rng in self:
            rst += rng.length()

        return rst

    def normalize(self):

        i = 0
        while i < len(self) - 1:
            curr = self[i]
            nxt = self[i+1]

            if not curr.is_adjacent(nxt):
                i += 1
                continue

            # compare value if there is
            if curr[2:] == nxt[2:]:
                o = [curr[0], nxt[1]] + curr[2:]
                self[i] = self.range_clz(*o)
                self.pop(i+1)
            else:
                i += 1
                continue

    def find_overlapped(self, rng):

        rng = list(rng)
        rst = []

        i = bisect_left(self, rng)
        while i < len(self):

            if self[i].cmp(rng) == 0:

                if self[i].intersect(rng) is not None:
                    rst.append(self.range_clz(*self[i]))
                else:
                    # adjacent ranges
                    pass
                i += 1
            else:
                break

        return self.__class__(rst)


class RangeSet(RangeDict):

    default_range_clz = Range

    def add(self, rng):

        rng = _to_range(self.range_clz, list(rng))

        i = bisect_left(self, rng)

        while i < len(self):
            if rng.cmp(self[i]) == 0:
                rng = union_range(rng, self[i])
                self.pop(i)
            else:
                break

        self.insert(i, rng)


class IntIncRangeSet(RangeSet):
    default_range_clz = IntIncRange


def union(a, *bs):
    u = a
    for b in bs:
        u = _union(u, b)
    return u


def _union(a, b):

    if len(a) == 0:
        return RangeSet(b, range_clz=b.range_clz)

    if len(b) == 0:
        return RangeSet(a, range_clz=b.range_clz)

    rst = []
    i, j = 1, 0

    rng = a[0]

    while i < len(a) or j < len(b):

        a_ge_b = None

        if i == len(a):
            a_ge_b = True
        elif j == len(b):
            a_ge_b = False
        else:
            a_ge_b = a[i].cmp_left(b[j][0]) >= 0

        if a_ge_b:
            nxt = b[j]
            j += 1
        else:
            nxt = a[i]
            i += 1

        if rng.cmp(nxt) == 0:
            rng = union_range(rng, nxt)
        else:
            assert rng.cmp(nxt) < 0
            rst.append(rng)
            rng = nxt

    rst.append(rng)

    return RangeSet(rst, range_clz=b.range_clz)


def substract(a, *bs):
    s = a
    for b in bs:
        s = _substract(s, b)
    return s


def _substract(a, b):

    if len(a) == 0:
        return a.__class__([], range_clz=a.range_clz)

    if len(b) == 0:
        return a.__class__(a, range_clz=a.range_clz)

    rst = []

    for ra in a:

        sb = a.range_clz(*ra)

        j = bisect_left(b, ra)

        while j < len(b) and sb.cmp(b[j]) == 0:
            u1, u2 = substract_range(sb, b[j])

            if u1 is not None:
                rst.append(u1)

            sb = u2

            if sb is None:
                break

            j += 1

        if sb is not None:
            rst.append(sb)

    return a.__class__(rst, range_clz=a.range_clz)


def intersect(a, b):
    return substract(a, substract(a, b))


def assert_type_valid(typ):
    if typ not in compatible_types:
        raise TypeError('{typ} is not comparable'.format(typ=typ))


def _to_range(range_clz, rng):

    # rangeset is 2 element iterable, rangedict is 3 element iterable
    if len(rng) < 2:
        raise ValueError('range length is at least 2 but {l}: {rng}'.format(
            l=len(rng), rng=rng))

    return range_clz(*rng)


def cmp_boundary(l, r):

    if l is None:
        # left is -inf
        return -1

    if r is None:
        # right is inf
        return -1

    if l < r:
        return -1
    elif l > r:
        return 1
    else:
        return 0


def cmp_val(a, b, none_cmp_finite=1):

    # compare two value. any of them can be None.
    # None means positive infinite or negative infinite, defined by none_cmp_finite

    if a is None:
        if b is None:
            return 0
        else:
            # compare(none, b)
            return none_cmp_finite
    else:
        if b is None:
            # compare(b, none) = -compare(none, b)
            return -none_cmp_finite
        else:
            if a < b:
                return -1
            elif a > b:
                return 1
            else:
                return 0


def union_range(a, b):

    if a.cmp(b) != 0:
        raise Unmergeable(a, b)

    if a[0] is None or b[0] is None:
        left = None
    else:
        left = min([a[0], b[0]])

    if a[1] is None or b[1] is None:
        right = None
    else:
        right = max([a[1], b[1]])

    return a.__class__(left, right)


def substract_range(a, b):

    if a.cmp(b) > 0:
        # keep value for ValueRange
        return [None,
                a.dup() + a[2:]]

    if a.cmp(b) < 0:
        # keep value for ValueRange
        return [a.dup() + a[2:],
                None]

    # keep value for ValueRange
    rst = [None, None]

    if a.cmp_left(b[0]) < 0:
        o = [a[0], b.prev_right()] + a[2:]
        rst[0] = a.__class__(*o)

    if b.cmp_right(a[1]) < 0:
        o = [b.next_left(), a[1]] + a[2:]
        rst[1] = a.__class__(*o)

    return rst


def bisect_left(a, x, lo=0, hi=None):

    # Find the left-most i so that a[i] >= x
    # Thus i is where to a.insert(i, x)

    if lo < 0:
        raise ValueError('lo must be non-negative')

    if hi is None:
        hi = len(a)

    while lo < hi:
        mid = (lo + hi) // 2
        if a[mid].cmp(x) < 0:
            lo = mid + 1
        else:
            hi = mid
    return lo


def assert_compatible(l, r):
    if not is_compatible(l, r):
        raise TypeError('{l} {ltyp} is incompatible with {r} {rtyp}'.format(
            l=l, ltyp=type(l), r=r, rtyp=type(r)))


def is_compatible(l, r):
    if l is None or r is None:
        return True

    if type(l) in compatible_types.get(type(r), ()):
        return True

    if type(r) in compatible_types.get(type(l), ()):
        return True

    return False



#!/usr/bin/env python2.6
# coding: utf-8

import unittest

from pykit import awssign


class TestSigner(unittest.TestCase):

    def test_basic(self):
        signer = awssign.Signer('access_key', 'secret_key')
        fields = {
            'key': 'test_key',
            'Policy': {
                'expiration': '2018-01-01T12:00:00.000Z',
                'condition': [
                    ['starts-with', '$key', ''],
                    {
                        'bucket': 'test-bucket',
                    },
                ],
            },
        }

        signer.add_post_auth(fields, request_date='20180101T120101Z')

        self.assertEqual('AWS4-HMAC-SHA256', fields['X-Amz-Algorithm'])
        self.assertEqual('20180101T120101Z', fields['X-Amz-Date'])
        self.assertEqual('19235d229144aa2a1d2b1c3a842c96dfb76bca9b75286c2a0aaae8e29f45c5a7',
                         fields['X-Amz-Signature'])
        self.assertEqual('access_key/20180101/us-east-1/s3/aws4_request',
                         fields['X-Amz-Credential'])
        self.assertEqual('eyJleHBpcmF0aW9uIjogIjIwMTgtMDEtMDFUMTI6MDA6MDAuMDAwWiIsICJjb25kaXRpb24iOiBbWyJzdGFydHMtd2l0aCIsICIka2V5IiwgIiJdLCB7ImJ1Y2tldCI6ICJ0ZXN0LWJ1Y2tldCJ9XX0=',
                         fields['Policy'])
